import logging
from datetime import datetime
from typing import Iterable, List, Optional

from hdx.location.adminone import AdminOne
from hdx.utilities.errors_onexit import ErrorsOnExit
from hdx.utilities.typehint import ListTuple

from .base_scraper import BaseScraper
from .configurable.scraper import ConfigurableScraper
from .utilities import get_isodate_from_dataset_date
from .utilities.fallbacks import Fallbacks
from .utilities.reader import Read

logger = logging.getLogger(__name__)


class Runner:
    """Runner class

    Args:
        countryiso3s (ListTuple[str]): List of ISO3 country codes to process
        adminone (AdminOne): AdminOne object from HDX Python Country library
        today (datetime): Value to use for today. Defaults to datetime.now().
        errors_on_exit (ErrorsOnExit): ErrorsOnExit object that logs errors on exit
        scrapers_to_run (Optional[ListTuple[str]]): Scrapers to run. Defaults to None.
    """

    def __init__(
        self,
        countryiso3s: ListTuple[str],
        adminone: AdminOne,
        today: datetime = datetime.now(),
        errors_on_exit: Optional[ErrorsOnExit] = None,
        scrapers_to_run: Optional[ListTuple[str]] = None,
    ):
        self.countryiso3s = countryiso3s
        self.adminone = adminone
        self.today = today
        self.errors_on_exit = errors_on_exit
        if isinstance(scrapers_to_run, tuple):
            scrapers_to_run = list(scrapers_to_run)
        self.scrapers_to_run: Optional[List[str]] = scrapers_to_run
        self.scrapers = dict()
        self.scraper_names = list()

    def add_custom(
        self, scraper: BaseScraper, add_to_run: bool = False
    ) -> str:
        scraper_name = scraper.name
        self.scrapers[scraper_name] = scraper
        if scraper_name not in self.scraper_names:
            self.scraper_names.append(scraper_name)
        if (
            add_to_run
            and self.scrapers_to_run is not None
            and scraper_name not in self.scrapers_to_run
        ):
            self.scrapers_to_run.append(scraper_name)
        scraper.errors_on_exit = self.errors_on_exit
        return scraper_name

    def add_customs(
        self, scrapers: Iterable[BaseScraper], add_to_run: bool = False
    ) -> List[str]:
        scraper_names = list()
        for scraper in scrapers:
            scraper_names.append(self.add_custom(scraper, add_to_run))
        return scraper_names

    def add_configurable(
        self, name, datasetinfo, level, level_name=None, suffix=None
    ) -> str:
        if suffix:
            key = f"{name}{suffix}"
        else:
            key = name
        if key not in self.scraper_names:
            self.scraper_names.append(key)
        self.scrapers[key] = ConfigurableScraper(
            name,
            datasetinfo,
            level,
            self.countryiso3s,
            self.adminone,
            level_name,
            self.today,
            self.errors_on_exit,
        )
        return key

    def add_configurables(
        self, configuration, level, level_name=None, suffix=None
    ) -> List[str]:
        keys = list()
        for name in configuration:
            datasetinfo = configuration[name]
            keys.append(
                self.add_configurable(
                    name, datasetinfo, level, level_name, suffix
                )
            )
        return keys

    def prioritise_scrapers(self, scraper_names):
        for scraper_name in reversed(scraper_names):
            if scraper_name in self.scraper_names:
                self.scraper_names.remove(scraper_name)
                self.scraper_names.insert(0, scraper_name)

    def get_scraper_names(self):
        return self.scraper_names

    def get_scraper(self, name):
        return self.scrapers.get(name)

    def get_scraper_exception(self, name):
        scraper = self.get_scraper(name)
        if not scraper:
            raise ValueError(f"No such scraper {name}!")
        return scraper

    def add_instance_variables(self, name, **kwargs):
        scraper = self.get_scraper_exception(name)
        for key, value in kwargs.items():
            setattr(scraper, key, value)

    def add_post_run(self, name, fn):
        scraper = self.get_scraper_exception(name)
        scraper.post_run = lambda: fn(scraper)

    def run_one(self, name, run_again=False):
        scraper = self.get_scraper_exception(name)
        if scraper.has_run is False or run_again:
            try:
                scraper.run()
                scraper.add_sources()
                scraper.add_source_urls()
                scraper.add_population()
                logger.info(f"Processed {scraper.name}")
            except Exception as ex:
                if not Fallbacks.exist() or scraper.can_fallback is False:
                    raise
                logger.exception(f"Using fallbacks for {scraper.name}!")
                if self.errors_on_exit:
                    self.errors_on_exit.add(
                        f"Using fallbacks for {scraper.name}! Error: {ex}"
                    )
                for level in scraper.headers.keys():
                    values, sources = Fallbacks.get(
                        level, scraper.headers[level]
                    )
                    scraper.values[level] = values
                    scraper.sources[level] = sources
                scraper.add_population()
                scraper.fallbacks_used = True
                scraper.run_after_fallbacks()
            scraper.has_run = True
            scraper.post_run()

    def run_scraper(self, name, run_again=False):
        if self.scrapers_to_run and not any(
            x in name for x in self.scrapers_to_run
        ):
            return False
        logger.info(f"Running {name}")
        self.run_one(name, run_again)
        return True

    def run(
        self, what_to_run=None, run_again=False, prioritise_scrapers=tuple()
    ):
        self.prioritise_scrapers(prioritise_scrapers)
        for name in self.scraper_names:
            if what_to_run and name not in what_to_run:
                continue
            self.run_scraper(name, run_again)

    def set_not_run(self, name):
        self.get_scraper(name).has_run = False

    def set_not_run_many(self, names):
        for name in names:
            self.get_scraper(name).has_run = False

    def get_headers(self, names=None, levels=None, headers=None, hxltags=None):
        if not names:
            names = self.scrapers.keys()
        results = dict()
        for name in names:
            if self.scrapers_to_run and not any(
                x in name for x in self.scrapers_to_run
            ):
                continue
            scraper = self.get_scraper(name)
            for level, scraper_headers in scraper.headers.items():
                if levels is not None and level not in levels:
                    continue
                level_results = results.get(level)
                if level_results is None:
                    level_results = (list(), list())
                    results[level] = level_results
                for i, header in enumerate(scraper_headers[0]):
                    if headers is not None and header not in headers:
                        continue
                    hxltag = scraper_headers[1][i]
                    if hxltags is not None and hxltag not in hxltags:
                        continue
                    level_results[0].append(header)
                    level_results[1].append(hxltag)
        return results

    def get_results(
        self, names=None, levels=None, overrides=dict(), has_run=True
    ):
        if not names:
            names = self.scrapers.keys()
        results = dict()

        def add_level_results(lvl, output_lvl, scrap, lvls_used):
            nonlocal results

            if level in lvls_used:
                return
            if levels is not None and output_lvl not in levels:
                return
            headers = scrap.headers.get(lvl)
            if headers is None:
                return
            level_results = results.get(output_lvl)
            if level_results is None:
                level_results = {
                    "headers": (list(), list()),
                    "values": list(),
                    "sources": list(),
                    "fallbacks": list(),
                }
                results[output_lvl] = level_results
            level_results["headers"][0].extend(headers[0])
            level_results["headers"][1].extend(headers[1])
            level_results["values"].extend(scrap.get_values(level))
            level_results["sources"].extend(scrap.get_sources(level))
            lvls_used.add(lvl)
            lvls_used.add(output_lvl)

        for name in names:
            if self.scrapers_to_run and not any(
                x in name for x in self.scrapers_to_run
            ):
                continue
            scraper = self.get_scraper(name)
            if has_run and not scraper.has_run:
                continue
            override = overrides.get(name, dict())
            levels_used = set()
            for level, output_level in override.items():
                add_level_results(level, output_level, scraper, levels_used)

            for level in scraper.headers:
                add_level_results(level, level, scraper, levels_used)

        return results

    def get_rows(
        self,
        level,
        adms,
        headers=(tuple(), tuple()),
        row_fns=tuple(),
        names=None,
        overrides=dict(),
    ):
        results = self.get_results(names, [level], overrides=overrides).get(
            level
        )
        rows = list()
        if results:
            all_headers = results["headers"]
            rows.append(list(headers[0]) + all_headers[0])
            rows.append(list(headers[1]) + all_headers[1])
            all_values = results["values"]
            for adm in adms:
                row = list()
                for fn in row_fns:
                    row.append(fn(adm))
                for values in all_values:
                    row.append(values.get(adm))
                rows.append(row)
        return rows

    def get_sources(self, names=None, levels=None, additional_sources=tuple()):
        if not names:
            names = self.scrapers.keys()
        hxltags = set()
        sources = list()
        reader = Read.get_reader()
        for sourceinfo in additional_sources:
            date = sourceinfo.get("source_date")
            if date is None:
                if sourceinfo.get("force_date_today", False):
                    date = self.today.strftime("%Y-%m-%d")
            source = sourceinfo.get("source")
            source_url = sourceinfo.get("source_url")
            dataset_name = sourceinfo.get("dataset")
            if dataset_name:
                dataset = reader.read_dataset(dataset_name)
                if date is None:
                    date = get_isodate_from_dataset_date(
                        dataset, today=self.today
                    )
                if source is None:
                    source = dataset["dataset_source"]
                if source_url is None:
                    source_url = dataset.get_hdx_url()
            sources.append((sourceinfo["indicator"], date, source, source_url))
        for name in names:
            if self.scrapers_to_run and not any(
                x in name for x in self.scrapers_to_run
            ):
                continue
            scraper = self.get_scraper(name)
            if not scraper.has_run:
                continue
            if levels is not None:
                levels_to_check = levels
            else:
                levels_to_check = scraper.sources.keys()
            for level in levels_to_check:
                for source in scraper.get_sources(level):
                    hxltag = source[0]
                    if hxltag in hxltags:
                        continue
                    hxltags.add(hxltag)
                    sources.append(source)
        return sources

    def get_source_urls(self, names=None):
        source_urls = set()
        if not names:
            names = self.scrapers.keys()
        for name in names:
            scraper = self.get_scraper(name)
            if not scraper.has_run:
                continue
            source_urls.update(scraper.get_source_urls())
        return sorted(source_urls)



from os.path import exists, join

import pytest
from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date
from hdx.utilities.path import temp_dir

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.configurable.filecopier import FileCopier
from hdx.scraper.runner import Runner


class TestScrapersFileCopier:
    @pytest.fixture(scope="function")
    def output_file(self, fixtures):
        return join(fixtures, "test_output.xlsx")

    def test_filecopier(self, configuration, output_file):
        with temp_dir(
            "TestScraperFileCopier",
            delete_on_success=True,
            delete_on_failure=False,
        ) as temp_folder:
            BaseScraper.population_lookup = dict()
            adminone = AdminOne(configuration)
            iso3s = ("AFG", "MMR")
            today = parse_date("2020-10-01")
            runner = Runner(iso3s, adminone, today)
            scrapers = FileCopier.get_scrapers(
                configuration["copyfiles"], temp_folder
            )
            scraper_names = runner.add_customs(scrapers, add_to_run=True)
            name = "filecopier_xlsx"
            assert scraper_names == [name]
            runner.run()
            assert runner.get_sources() == [
                (
                    "#xlsx",
                    "2022-04-14",
                    "Multiple Sources",
                    "https://data.humdata.org/dataset/ukraine-border-crossings",
                )
            ]
            assert exists(output_file)



import logging

import pytest
from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.runner import Runner

from .conftest import run_check_scraper


class TestScraperGlobal:
    @pytest.fixture(scope="class")
    def cerf_headers(self):
        return (
            [
                "CBPFFunding",
                "CBPFFundingGMEmpty",
                "CBPFFundingGM0",
                "CBPFFundingGM1",
                "CBPFFundingGM2",
                "CBPFFundingGM3",
                "CBPFFundingGM4",
                "CERFFunding",
                "CERFFundingGMEmpty",
                "CERFFundingGM0",
                "CERFFundingGM1",
                "CERFFundingGM2",
                "CERFFundingGM3",
                "CERFFundingGM4",
            ],
            [
                "#value+cbpf+funding+total+usd",
                "#value+cbpf+funding+gmempty+total+usd",
                "#value+cbpf+funding+gm0+total+usd",
                "#value+cbpf+funding+gm1+total+usd",
                "#value+cbpf+funding+gm2+total+usd",
                "#value+cbpf+funding+gm3+total+usd",
                "#value+cbpf+funding+gm4+total+usd",
                "#value+cerf+funding+total+usd",
                "#value+cerf+funding+gmempty+total+usd",
                "#value+cerf+funding+gm0+total+usd",
                "#value+cerf+funding+gm1+total+usd",
                "#value+cerf+funding+gm2+total+usd",
                "#value+cerf+funding+gm3+total+usd",
                "#value+cerf+funding+gm4+total+usd",
            ],
        )

    def test_get_global_2020(self, configuration, cerf_headers):
        BaseScraper.population_lookup = dict()
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)
        level = "single"
        level_name = "global"
        scraper_configuration = configuration[f"scraper_{level_name}"]
        runner = Runner(configuration["HRPs"], adminone, today)
        keys = runner.add_configurables(
            scraper_configuration, level, level_name
        )
        assert keys == [
            "covax",
            "ourworldindata",
            "cerf_global",
            "broken_cerf_url",
        ]

        name = "covax"
        headers = (
            [
                "Covax Interim Forecast Doses",
                "Covax Delivered Doses",
                "Other Delivered Doses",
                "Total Delivered Doses",
                "Covax Pfizer-BioNTech Doses",
                "Covax Astrazeneca-SII Doses",
                "Covax Astrazeneca-SKBio Doses",
            ],
            [
                "#capacity+doses+forecast+covax",
                "#capacity+doses+delivered+covax",
                "#capacity+doses+delivered+others",
                "#capacity+doses+delivered+total",
                "#capacity+doses+covax+pfizerbiontech",
                "#capacity+doses+covax+astrazenecasii",
                "#capacity+doses+covax+astrazenecaskbio",
            ],
        )
        values = [
            {"value": "73248240"},
            {"value": "12608040"},
            {"value": "23728358"},
            {"value": "36336398"},
            {"value": "271440"},
            {"value": "67116000"},
            {"value": "5860800"},
        ]
        sources = [
            (
                "#capacity+doses+forecast+covax",
                "2020-08-07",
                "covax",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vTVzu79PPTfaA2syevOQfyRRjy63dJWitqu0fFbXIQCzoUn9K9TiMWMRvFGg1RBsnLmgYugzSEiAye2/pub?gid=992438980&single=true&output=csv",
            ),
            (
                "#capacity+doses+delivered+covax",
                "2020-08-07",
                "covax",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vTVzu79PPTfaA2syevOQfyRRjy63dJWitqu0fFbXIQCzoUn9K9TiMWMRvFGg1RBsnLmgYugzSEiAye2/pub?gid=992438980&single=true&output=csv",
            ),
            (
                "#capacity+doses+delivered+others",
                "2020-08-07",
                "covax",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vTVzu79PPTfaA2syevOQfyRRjy63dJWitqu0fFbXIQCzoUn9K9TiMWMRvFGg1RBsnLmgYugzSEiAye2/pub?gid=992438980&single=true&output=csv",
            ),
            (
                "#capacity+doses+delivered+total",
                "2020-08-07",
                "covax",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vTVzu79PPTfaA2syevOQfyRRjy63dJWitqu0fFbXIQCzoUn9K9TiMWMRvFGg1RBsnLmgYugzSEiAye2/pub?gid=992438980&single=true&output=csv",
            ),
            (
                "#capacity+doses+covax+pfizerbiontech",
                "2020-08-07",
                "covax",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vTVzu79PPTfaA2syevOQfyRRjy63dJWitqu0fFbXIQCzoUn9K9TiMWMRvFGg1RBsnLmgYugzSEiAye2/pub?gid=992438980&single=true&output=csv",
            ),
            (
                "#capacity+doses+covax+astrazenecasii",
                "2020-08-07",
                "covax",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vTVzu79PPTfaA2syevOQfyRRjy63dJWitqu0fFbXIQCzoUn9K9TiMWMRvFGg1RBsnLmgYugzSEiAye2/pub?gid=992438980&single=true&output=csv",
            ),
            (
                "#capacity+doses+covax+astrazenecaskbio",
                "2020-08-07",
                "covax",
                "https://docs.google.com/spreadsheets/d/e/2PACX-1vTVzu79PPTfaA2syevOQfyRRjy63dJWitqu0fFbXIQCzoUn9K9TiMWMRvFGg1RBsnLmgYugzSEiAye2/pub?gid=992438980&single=true&output=csv",
            ),
        ]
        run_check_scraper(
            name,
            runner,
            level_name,
            headers,
            values,
            sources,
            set_not_run=False,
        )
        rows = runner.get_rows("global", ("value",))
        expected_rows = [
            [
                "Covax Interim Forecast Doses",
                "Covax Delivered Doses",
                "Other Delivered Doses",
                "Total Delivered Doses",
                "Covax Pfizer-BioNTech Doses",
                "Covax Astrazeneca-SII Doses",
                "Covax Astrazeneca-SKBio Doses",
            ],
            [
                "#capacity+doses+forecast+covax",
                "#capacity+doses+delivered+covax",
                "#capacity+doses+delivered+others",
                "#capacity+doses+delivered+total",
                "#capacity+doses+covax+pfizerbiontech",
                "#capacity+doses+covax+astrazenecasii",
                "#capacity+doses+covax+astrazenecaskbio",
            ],
            [
                "73248240",
                "12608040",
                "23728358",
                "36336398",
                "271440",
                "67116000",
                "5860800",
            ],
        ]
        assert rows == expected_rows
        rows = runner.get_rows(
            "gho", ("value",), overrides={name: {"global": "gho"}}
        )
        assert rows == expected_rows
        runner.set_not_run(name)

        name = "cerf_global"
        headers = cerf_headers
        values = [
            {"value": 906790749.5500005},
            {"value": 829856355.4100008},
            {"value": 37432868.04999999},
            {"value": 39501526.08999999},
            {},
            {},
            {},
            {"value": 848145238.0},
            {},
            {"value": 50042305.0},
            {"value": 75349572.0},
            {"value": 224560378.0},
            {"value": 349338181.0},
            {"value": 147855321.0},
        ]
        sources = [
            (
                "#value+cbpf+funding+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gmempty+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm0+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm1+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm2+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm3+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm4+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gmempty+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm0+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm1+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm2+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm3+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm4+total+usd",
                "2020-10-01",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
        ]
        run_check_scraper(name, runner, level_name, headers, values, sources)

        name = "ourworldindata"
        headers = (
            ["TotalDosesAdministered"],
            ["#capacity+doses+administered+total"],
        )
        values = [dict()]
        sources = [
            (
                "#capacity+doses+administered+total",
                "2020-10-01",
                "Our World in Data",
                "https://proxy.hxlstandard.org/data.csv?tagger-match-all=on&tagger-01-header=location&tagger-01-tag=%23country%2Bname&tagger-02-header=iso_code&tagger-02-tag=%23country%2Bcode&tagger-03-header=date&tagger-03-tag=%23date&tagger-04-header=total_vaccinations&tagger-04-tag=%23total%2Bvaccinations&tagger-08-header=daily_vaccinations&tagger-08-tag=%23total%2Bvaccinations%2Bdaily&url=https%3A%2F%2Fraw.githubusercontent.com%2Fowid%2Fcovid-19-data%2Fmaster%2Fpublic%2Fdata%2Fvaccinations%2Fvaccinations.csv&header-row=1&dest=data_view",
            )
        ]
        run_check_scraper(name, runner, level_name, headers, values, sources)

    def test_get_global_2021(
        self, cerf_headers, caplog, configuration, fallbacks
    ):
        BaseScraper.population_lookup = dict()
        today = parse_date("2021-05-03")
        adminone = AdminOne(configuration)
        level = "single"
        level_name = "global"
        scraper_configuration = configuration[f"scraper_{level_name}"]
        runner = Runner(configuration["HRPs"], adminone, today)
        runner.add_configurables(scraper_configuration, level, level_name)
        name = "cerf_global"
        headers = cerf_headers
        values = [
            {"value": 7811774.670000001},
            {"value": 7811774.670000001},
            {},
            {},
            {},
            {},
            {},
            {"value": 89298919.0},
            {"value": 6747034.0},
            {},
            {"value": 2549855.0},
            {"value": 10552572.0},
            {"value": 26098816.0},
            {"value": 43350642.0},
        ]
        sources = [
            (
                "#value+cbpf+funding+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gmempty+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm0+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm1+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm2+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm3+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cbpf+funding+gm4+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gmempty+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm0+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm1+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm2+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm3+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
            (
                "#value+cerf+funding+gm4+total+usd",
                "2021-05-03",
                "CERF and CBPF",
                "https://data.humdata.org/dataset/cerf-covid-19-allocations",
            ),
        ]
        run_check_scraper(name, runner, level_name, headers, values, sources)

        # Test fallbacks with subsets
        with caplog.at_level(logging.ERROR):
            name = "broken_cerf_url"
            headers = cerf_headers
            values = [
                {"value": 7811775.670000001},
                {"value": 7811775.670000001},
                {},
                {},
                {},
                {},
                {},
                {"value": 89298924.0},
                {"value": 6747035.0},
                {},
                {"value": 2549856.0},
                {"value": 10552573.0},
                {"value": 26098817.0},
                {"value": 43350643.0},
            ]
            sources = [
                (
                    "#value+cbpf+funding+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cbpf+funding+gmempty+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cbpf+funding+gm0+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cbpf+funding+gm1+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cbpf+funding+gm2+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cbpf+funding+gm3+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cbpf+funding+gm4+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cerf+funding+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cerf+funding+gmempty+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cerf+funding+gm0+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cerf+funding+gm1+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cerf+funding+gm2+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cerf+funding+gm3+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
                (
                    "#value+cerf+funding+gm4+total+usd",
                    "2020-09-01",
                    "CERF and CBPF",
                    "tests/fixtures/fallbacks.json",
                ),
            ]
            run_check_scraper(
                name,
                runner,
                level_name,
                headers,
                values,
                sources,
                fallbacks_used=True,
            )
            assert f"Using fallbacks for {name}!" in caplog.text
            assert (
                "No such file or directory: 'tests/fixtures/broken_cerf_url_notexist.csv'"
                in caplog.text
            )

        name = "ourworldindata"
        headers = (
            ["TotalDosesAdministered"],
            ["#capacity+doses+administered+total"],
        )
        values = [{"value": "13413871"}]
        sources = [
            (
                "#capacity+doses+administered+total",
                "2021-05-03",
                "Our World in Data",
                "https://proxy.hxlstandard.org/data.csv?tagger-match-all=on&tagger-01-header=location&tagger-01-tag=%23country%2Bname&tagger-02-header=iso_code&tagger-02-tag=%23country%2Bcode&tagger-03-header=date&tagger-03-tag=%23date&tagger-04-header=total_vaccinations&tagger-04-tag=%23total%2Bvaccinations&tagger-08-header=daily_vaccinations&tagger-08-tag=%23total%2Bvaccinations%2Bdaily&url=https%3A%2F%2Fraw.githubusercontent.com%2Fowid%2Fcovid-19-data%2Fmaster%2Fpublic%2Fdata%2Fvaccinations%2Fvaccinations.csv&header-row=1&dest=data_view",
            )
        ]
        run_check_scraper(name, runner, level_name, headers, values, sources)

        scraper_configuration = configuration["other"]
        runner.add_configurables(scraper_configuration, level, level_name)
        name = "population_other"
        headers = (
            ["Population"],
            ["#population"],
        )
        val = 7673533972
        values = [{"value": val}]
        pop_value = {"global": val}
        sources = [
            (
                "#population",
                "2021-05-03",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            )
        ]
        run_check_scraper(
            name,
            runner,
            level_name,
            headers,
            values,
            sources,
            population_lookup=pop_value,
        )

        name = "ourworldindata_other"
        headers = (
            ["TotalDosesAdministered"],
            ["#capacity+doses+administered+total"],
        )
        values = [{"value": "1175451507"}]
        sources = [
            (
                "#capacity+doses+administered+total",
                "2021-05-03",
                "Our World in Data",
                "https://proxy.hxlstandard.org/data.csv?tagger-match-all=on&tagger-01-header=location&tagger-01-tag=%23country%2Bname&tagger-02-header=iso_code&tagger-02-tag=%23country%2Bcode&tagger-03-header=date&tagger-03-tag=%23date&tagger-04-header=total_vaccinations&tagger-04-tag=%23total%2Bvaccinations&tagger-08-header=daily_vaccinations&tagger-08-tag=%23total%2Bvaccinations%2Bdaily&url=https%3A%2F%2Fraw.githubusercontent.com%2Fowid%2Fcovid-19-data%2Fmaster%2Fpublic%2Fdata%2Fvaccinations%2Fvaccinations.csv&header-row=1&dest=data_view",
            )
        ]
        run_check_scraper(name, runner, level_name, headers, values, sources)

        name = "altworldindata"
        headers = (
            [
                "TotalDosesAdministered",
                "PopulationCoverageAdministeredDoses",
            ],
            [
                "#capacity+doses+administered+total",
                "#capacity+doses+administered+coverage+pct",
            ],
        )
        values = [{"value": 1175451507}, {"value": "0.0766"}]
        sources = [
            (
                "#capacity+doses+administered+total",
                "2021-05-03",
                "Our World in Data",
                "https://proxy.hxlstandard.org/data.csv?tagger-match-all=on&tagger-01-header=location&tagger-01-tag=%23country%2Bname&tagger-02-header=iso_code&tagger-02-tag=%23country%2Bcode&tagger-03-header=date&tagger-03-tag=%23date&tagger-04-header=total_vaccinations&tagger-04-tag=%23total%2Bvaccinations&tagger-08-header=daily_vaccinations&tagger-08-tag=%23total%2Bvaccinations%2Bdaily&url=https%3A%2F%2Fraw.githubusercontent.com%2Fowid%2Fcovid-19-data%2Fmaster%2Fpublic%2Fdata%2Fvaccinations%2Fvaccinations.csv&header-row=1&dest=data_view",
            ),
            (
                "#capacity+doses+administered+coverage+pct",
                "2021-05-03",
                "Our World in Data",
                "https://proxy.hxlstandard.org/data.csv?tagger-match-all=on&tagger-01-header=location&tagger-01-tag=%23country%2Bname&tagger-02-header=iso_code&tagger-02-tag=%23country%2Bcode&tagger-03-header=date&tagger-03-tag=%23date&tagger-04-header=total_vaccinations&tagger-04-tag=%23total%2Bvaccinations&tagger-08-header=daily_vaccinations&tagger-08-tag=%23total%2Bvaccinations%2Bdaily&url=https%3A%2F%2Fraw.githubusercontent.com%2Fowid%2Fcovid-19-data%2Fmaster%2Fpublic%2Fdata%2Fvaccinations%2Fvaccinations.csv&header-row=1&dest=data_view",
            ),
        ]
        run_check_scraper(name, runner, level_name, headers, values, sources)



import copy
import logging
from datetime import datetime
from typing import Any, Dict, Iterator, List, Optional, Tuple

import regex
from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import get_datetime_from_timestamp, parse_date
from hdx.utilities.dictandlist import dict_of_lists_add
from hdx.utilities.downloader import DownloadError
from hdx.utilities.errors_onexit import ErrorsOnExit
from hdx.utilities.text import (  # noqa: F401
    get_fraction_str,
    get_numeric_if_possible,
    number_format,
)

from ..base_scraper import BaseScraper
from ..utilities import get_rowval
from .rowparser import RowParser

logger = logging.getLogger(__name__)


class ConfigurableScraper(BaseScraper):
    """Runs one mini scraper given dataset information and returns headers, values and
    sources.

    Args:
        name (str): Name of scraper
        datasetinfo (Dict): Information about dataset
        level (str): Can be national, subnational or single
        countryiso3s (List[str]): List of ISO3 country codes to process
        adminone (AdminOne): AdminOne object from HDX Python Country library
        level_name (Optional[str]): Customised level_name name. Defaults to None (level_name).
        today (datetime): Value to use for today. Defaults to datetime.now().
        errors_on_exit (Optional[ErrorsOnExit]): ErrorsOnExit object that logs errors on exit
        **kwargs: Variables to use when evaluating template arguments in urls
    """

    brackets = r"""
    (?<rec> #capturing group rec
     \( #open parenthesis
     (?: #non-capturing group
      [^()]++ #anything but parenthesis one or more times without backtracking
      | #or
       (?&rec) #recursive substitute of group rec
     )*
     \) #close parenthesis
    )"""

    def __init__(
        self,
        name: str,
        datasetinfo: Dict,
        level: str,
        countryiso3s: List[str],
        adminone: AdminOne,
        level_name: Optional[str] = None,
        today: datetime = datetime.now(),
        errors_on_exit: Optional[ErrorsOnExit] = None,
        **kwargs: Any,
    ):
        self.level = level
        datelevel = datasetinfo.get("date_level")
        if datelevel is None:
            self.datelevel = self.level
        else:
            self.datelevel = datelevel
        if level_name is None:
            self.level_name: str = level
        else:
            self.level_name: str = level_name
        self.countryiso3s = countryiso3s
        self.adminone = adminone
        self.today = today
        self.subsets = self.get_subsets_from_datasetinfo(datasetinfo)
        self.errors_on_exit: Optional[ErrorsOnExit] = errors_on_exit
        self.variables = kwargs
        self.rowparser = None
        self.datasetinfo = copy.deepcopy(datasetinfo)
        self.use_hxl_called = False
        headers = {self.level_name: (list(), list())}
        for subset in self.subsets:
            headers[self.level_name][0].extend(subset["output"])
            headers[self.level_name][1].extend(subset["output_hxl"])
        self.can_fallback = True
        if len(headers[self.level_name][0]) == 0:
            use_hxl = self.datasetinfo.get("use_hxl", False)
            if use_hxl:
                try:
                    file_headers, iterator = self.get_iterator(name)
                    self.use_hxl(headers, file_headers, iterator)
                except DownloadError:
                    self.can_fallback = False
        self.setup(name, headers)

    @staticmethod
    def get_subsets_from_datasetinfo(datasetinfo) -> List[Dict]:
        """Get subsets from dataset information

        Returns:
            List[Dict]: List of subsets
        """
        subsets = datasetinfo.get("subsets")
        if not subsets:
            subsets = [
                {
                    "filter": datasetinfo.get("filter"),
                    "input": datasetinfo.get("input", list()),
                    "transform": datasetinfo.get("transform", dict()),
                    "population_key": datasetinfo.get("population_key"),
                    "process": datasetinfo.get("process", list()),
                    "input_keep": datasetinfo.get("input_keep", list()),
                    "input_append": datasetinfo.get("input_append", list()),
                    "sum": datasetinfo.get("sum"),
                    "input_ignore_vals": datasetinfo.get(
                        "input_ignore_vals", list()
                    ),
                    "output": datasetinfo.get("output", list()),
                    "output_hxl": datasetinfo.get("output_hxl", list()),
                }
            ]
        return subsets

    def get_iterator(self, name):
        return self.get_reader(name).read(
            self.datasetinfo,
            **self.variables,
        )

    def add_sources(
        self,
    ) -> List[Tuple]:
        """Get source for each HXL hashtag

        Returns:
            List[Tuple]: List of (hxltag, date, source, source url)
        """
        date = self.datasetinfo.get("source_date")
        use_date_from_date_col = self.datasetinfo.get("use_date", False)
        if not date or use_date_from_date_col:
            date = self.rowparser.get_maxdate()
            if date == 0:
                raise ValueError(
                    "No date given in datasetinfo or as a column!"
                )
            if self.rowparser.datetype == "date":
                if not isinstance(date, datetime):
                    date = parse_date(date)
            elif self.rowparser.datetype == "int":
                date = get_datetime_from_timestamp(date)
            else:
                raise ValueError("No date type specified!")
        self.datasetinfo["source_date"] = date
        return super().add_sources()

    def add_population(self):
        return

    def read_hxl(self, iterator):
        use_hxl = self.datasetinfo.get("use_hxl", False)
        if not use_hxl:
            return None
        header_to_hxltag = next(iterator)
        while not header_to_hxltag:
            header_to_hxltag = next(iterator)
        return header_to_hxltag

    def use_hxl(
        self, headers, file_headers: List[str], iterator: Iterator[Dict]
    ) -> Optional[Dict]:
        """If the mini scraper configuration defines that HXL is used (use_hxl is True),
        then read the mapping from headers to HXL hash tags. Since each row is a
        dictionary from header to value, the HXL row will be a dictionary from header
        to HXL hashtag. Label #country and #adm1 columns as admin columns. If the
        input columns to use are not specified, use all that have HXL hashtags. If the
        output column headers or hashtags are not specified, use the ones from the
        original file.

        Args:
            file_headers (List[str]): List of all headers of input file
            iterator (Iterator[Dict]): Iterator over the rows

        Returns:
            Optional[Dict]: Dictionary that maps from header to HXL hashtag or None
        """
        header_to_hxltag = self.read_hxl(iterator)
        if not header_to_hxltag:
            return None
        if self.use_hxl_called:
            return header_to_hxltag
        self.use_hxl_called = True
        exclude_tags = self.datasetinfo.get("exclude_tags", list())
        find_tags = self.datasetinfo.get("find_tags")
        adm_cols = list()
        input_cols = list()
        columns = list()
        for header in file_headers:
            hxltag = header_to_hxltag[header]
            if not hxltag or hxltag in exclude_tags:
                continue
            if find_tags or (find_tags is None and self.datelevel != "single"):
                if "#country" in hxltag:
                    if "code" in hxltag:
                        if len(adm_cols) == 0:
                            adm_cols.append(hxltag)
                        else:
                            adm_cols[0] = hxltag
                    continue
                if find_tags or self.datelevel != "national":
                    if "#adm1" in hxltag:
                        if "code" in hxltag:
                            if len(adm_cols) == 0:
                                adm_cols.append(None)
                            if len(adm_cols) == 1:
                                adm_cols.append(hxltag)
                        continue
            if (
                hxltag == self.datasetinfo.get("date")
                and self.datasetinfo.get("include_date", False) is False
            ):
                continue
            input_cols.append(hxltag)
            columns.append(header)
        self.datasetinfo["admin"] = adm_cols
        for subset in self.subsets:
            orig_input_cols = subset.get("input", list())
            if not orig_input_cols:
                orig_input_cols.extend(input_cols)
            subset["input"] = orig_input_cols
            orig_columns = subset.get("output", list())
            if not orig_columns:
                orig_columns.extend(columns)
            subset["output"] = orig_columns
            orig_hxltags = subset.get("output_hxl", list())
            if not orig_hxltags:
                orig_hxltags.extend(input_cols)
            subset["output_hxl"] = orig_hxltags
            if headers:
                headers[self.level_name][0].extend(orig_columns)
                headers[self.level_name][1].extend(orig_hxltags)
        return header_to_hxltag

    def run_scraper(self, iterator: Iterator[Dict]) -> None:
        """Run one mini scraper given an iterator over the rows

        Args:
            iterator (Iterator[Dict]): Iterator over the rows

        Returns:
            None
        """

        valuedicts = dict()
        for subset in self.subsets:
            for _ in subset["input"]:
                dict_of_lists_add(valuedicts, subset["filter"], dict())

        def add_row(row):
            adm, should_process_subset = self.rowparser.parse(row)
            if not adm:
                return
            for i, subset in enumerate(self.subsets):
                if not should_process_subset[i]:
                    continue
                filter = subset["filter"]
                input_ignore_vals = subset.get("input_ignore_vals", list())
                input_transforms = subset.get("transform", dict())
                sum_cols = subset.get("sum")
                process_cols = subset.get("process")
                input_append = subset.get("input_append", list())
                input_keep = subset.get("input_keep", list())
                for i, valcol in enumerate(subset["input"]):
                    valuedict = valuedicts[filter][i]
                    val = get_rowval(row, valcol)
                    input_transform = input_transforms.get(valcol)
                    if input_transform and val not in input_ignore_vals:
                        val = eval(input_transform.replace(valcol, "val"))
                    if sum_cols or process_cols:
                        dict_of_lists_add(valuedict, adm, val)
                    else:
                        curval = valuedict.get(adm)
                        if valcol in input_append:
                            if curval:
                                val = curval + val
                        elif valcol in input_keep:
                            if curval:
                                val = curval
                        valuedict[adm] = val

        for row in self.rowparser.filter_sort_rows(iterator):
            add_row(row)

        values = self.values[self.level_name]
        values_pos = 0
        for subset in self.subsets:
            valdicts = valuedicts[subset["filter"]]
            population_key = subset.get("population_key")
            if population_key is None:
                population_str = "self.population_lookup[adm]"
            else:
                population_str = "self.population_lookup[population_key]"
            process_cols = subset.get("process")
            input_keep = subset.get("input_keep", list())
            sum_cols = subset.get("sum")
            input_ignore_vals = subset.get("input_ignore_vals", list())
            valcols = subset["input"]
            output_hxltags = subset["output_hxl"]
            # Indices of list sorted by length
            sorted_len_indices = sorted(
                range(len(valcols)),
                key=lambda k: len(valcols[k]),
                reverse=True,
            )

            def add_population(index, adm, value):
                if not value:
                    return
                if output_hxltags[index] == "#population":
                    if population_key is None:
                        self.population_lookup[adm] = value
                    else:
                        self.population_lookup[population_key] = value

            if process_cols:

                def text_replacement(string, adm):
                    # pzbgvjh is arbitrary! It is simply to prevent accidental replacement
                    # of all or parts of #population (if it is in the string).
                    arbitrary_string = "#pzbgvjh"
                    string = string.replace("#population", arbitrary_string)
                    hasvalues = False
                    for j in sorted_len_indices:
                        valcol = valcols[j]
                        if valcol not in string:
                            continue
                        if valcol in input_keep:
                            input_keep_index = 0
                        else:
                            input_keep_index = -1
                        val = valdicts[j][adm][input_keep_index]
                        if (
                            val is None
                            or val == ""
                            or val in input_ignore_vals
                        ):
                            val = 0
                        else:
                            hasvalues = True
                        string = string.replace(valcol, str(val))
                    string = string.replace(arbitrary_string, "#population")
                    return string, hasvalues

                for i, process_col in enumerate(process_cols):
                    valdict0 = valdicts[0]
                    for adm in valdict0:
                        hasvalues = True
                        matches = regex.search(
                            self.brackets, process_col, flags=regex.VERBOSE
                        )
                        if matches:
                            for bracketed_str in matches.captures("rec"):
                                if any(bracketed_str in x for x in valcols):
                                    continue
                                _, hasvalues_t = text_replacement(
                                    bracketed_str, adm
                                )
                                if not hasvalues_t:
                                    hasvalues = False
                                    break
                        if hasvalues:
                            formula, hasvalues_t = text_replacement(
                                process_col, adm
                            )
                            if hasvalues_t:
                                formula = formula.replace(
                                    "#population",
                                    population_str,
                                )
                                value = eval(formula)
                            else:
                                value = ""
                        else:
                            value = ""
                        values[values_pos][adm] = value
                        add_population(i, adm, value)
                    values_pos += 1
            elif sum_cols:
                for ind, sum_col in enumerate(sum_cols):
                    formula = sum_col["formula"]
                    mustbepopulated = sum_col.get("mustbepopulated", False)
                    newvaldicts = [dict() for _ in valdicts]
                    valdict0 = valdicts[0]
                    for adm in valdict0:
                        for i, val in enumerate(valdict0[adm]):
                            if not val or val in input_ignore_vals:
                                exists = False
                            else:
                                exists = True
                                for valdict in valdicts[1:]:
                                    val = valdict[adm][i]
                                    if (
                                        val is None
                                        or val == ""
                                        or val in input_ignore_vals
                                    ):
                                        exists = False
                                        break
                            if mustbepopulated and not exists:
                                continue
                            for j, valdict in enumerate(valdicts):
                                val = valdict[adm][i]
                                if (
                                    val is None
                                    or val == ""
                                    or val in input_ignore_vals
                                ):
                                    continue
                                newvaldicts[j][adm] = eval(
                                    f"newvaldicts[j].get(adm, 0.0) + {str(valdict[adm][i])}"
                                )
                    formula = formula.replace("#population", "#pzbgvjh")
                    for i in sorted_len_indices:
                        formula = formula.replace(
                            valcols[i], f"newvaldicts[{i}][adm]"
                        )
                    formula = formula.replace("#pzbgvjh", population_str)
                    for adm in valdicts[0].keys():
                        try:
                            val = eval(formula)
                        except (ValueError, TypeError, KeyError):
                            val = ""
                        values[values_pos][adm] = val
                        add_population(ind, adm, val)
                    values_pos += 1
            else:
                for i, valdict in enumerate(valdicts):
                    for adm in valdict:
                        value = valdict[adm]
                        values[values_pos][adm] = value
                        add_population(i, adm, value)
                    values_pos += 1

    def run(self) -> None:
        """Runs one mini scraper given dataset information

        Returns:
            None
        """
        file_headers, iterator = self.get_iterator(self.name)
        header_to_hxltag = self.use_hxl(None, file_headers, iterator)
        if "source_url" not in self.datasetinfo:
            self.datasetinfo["source_url"] = self.datasetinfo["url"]
        date = self.datasetinfo.get("source_date")
        if date:
            if isinstance(date, str):
                self.datasetinfo["source_date"] = parse_date(date)
        if not date or self.datasetinfo.get("force_date_today", False):
            self.datasetinfo["source_date"] = self.today
        self.rowparser = RowParser(
            self.name,
            self.countryiso3s,
            self.adminone,
            self.level,
            self.datelevel,
            self.datasetinfo,
            file_headers,
            header_to_hxltag,
            self.subsets,
        )
        self.run_scraper(iterator)



import logging
from typing import Dict, List, Optional, Union

from openpyxl import Workbook
from openpyxl.utils.dataframe import dataframe_to_rows

from .base import BaseOutput

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None


logger = logging.getLogger(__name__)


class ExcelFile(BaseOutput):
    """ExcelFile class enabling writing to Excel spreadsheets.

    Args:
        excel_path (str): Path to output spreadsheet
        tabs (Dict[str, str]): Dictionary of mappings from internal name to spreadsheet tab name
        updatetabs (List[str]): Tabs to update
    """

    def __init__(
        self, excel_path: str, tabs: Dict[str, str], updatetabs: List[str]
    ) -> None:
        super().__init__(updatetabs)
        self.workbook = Workbook()
        self.excel_path = excel_path
        self.tabs = tabs

    def update_tab(
        self,
        tabname: str,
        values: Union[List, DataFrame],
        hxltags: Optional[Dict] = None,
    ) -> None:
        """Update tab with values

        Args:
            tabname (str): Tab to update
            values (Union[List, DataFrame]): Values in a list of lists or a DataFrame
            hxltags (Optional[Dict]): HXL tag mapping. Defaults to None.

        Returns:
            None
        """
        if tabname not in self.updatetabs:
            return
        sheetname = self.tabs[tabname]
        try:
            del self.workbook[sheetname]
        except KeyError:
            pass
        tab = self.workbook.create_sheet(sheetname)
        if isinstance(values, list):
            for i, row in enumerate(values):
                for j, value in enumerate(row):
                    tab.cell(row=i + 1, column=j + 1, value=value)
        else:
            headers = list(values.columns.values)
            tab.append(headers)
            if hxltags:
                tab.append([hxltags.get(header, "") for header in headers])
            for r in dataframe_to_rows(values, index=False, header=False):
                tab.append(r)

    def save(self) -> None:
        """Save spreadsheet

        Returns:
            None
        """
        self.workbook.save(self.excel_path)



import logging

logger = logging.getLogger(__name__)


def idps_post_run(self) -> None:
    try:
        url = self.overrideinfo["url"]
        retriever = self.get_reader(prefix="idps_override")
        json = retriever.download_json(url)
        number_idps = int(json["data"][0]["individuals"])
        index = self.get_headers("national")[1].index("#affected+displaced")
        values = self.get_values("national")[index]
        for key, current_idps in values.items():
            if key != "MMR":
                continue
            logger.info(
                f"Replacing {current_idps} with {number_idps} for MMR IDPs!"
            )
            values[key] = number_idps
            self.get_source_urls().add(url)
            logger.info("Processed UNHCR Myanmar IDPs")
            break
    except Exception as ex:
        msg = "Not using UNHCR Myanmar IDPs override!"
        logger.exception(msg)
        if self.errors_on_exit:
            self.errors_on_exit.add(f"{msg} Error: {ex}")



"""Global fixtures"""
from os.path import join

import pytest
from hdx.api.configuration import Configuration
from hdx.api.locations import Locations
from hdx.location.country import Country
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.utilities import string_params_to_dict
from hdx.scraper.utilities.fallbacks import Fallbacks
from hdx.scraper.utilities.reader import Read

from . import bool_assert


@pytest.fixture(scope="session")
def fixtures():
    return join("tests", "fixtures")


@pytest.fixture(scope="session")
def configuration(fixtures):
    Configuration._create(
        hdx_read_only=True,
        hdx_site="prod",
        user_agent="test",
        project_config_yaml=join(
            "tests", "config", "project_configuration.yml"
        ),
    )
    Locations.set_validlocations(
        [
            {"name": "afg", "title": "Afghanistan"},
            {"name": "phl", "title": "Philippines"},
            {"name": "pse", "title": "State of Palestine"},
        ]
    )
    Country.countriesdata(use_live=False)

    header_auths = "population:pop_12345,who_national:who_abc"
    basic_auths = "access:YWNjXzEyMzQ1OmFjY19hYmM=,who_national2:d2hvX2RlZjp3aG9fMTIzNDU="
    param_auths = (
        "sadd:user=sadd_123&pass=sadd_abc,ourworldindata:auth=owid_abc"
    )

    header_auths = string_params_to_dict(header_auths)
    basic_auths = string_params_to_dict(basic_auths)
    param_auths = string_params_to_dict(param_auths)
    today = parse_date("2020-10-01")
    save = False
    if save:
        Read.create_readers(
            "",
            join(fixtures, "tmp"),
            "",
            save=True,
            use_saved=False,
            user_agent="test",
            header_auths=header_auths,
            basic_auths=basic_auths,
            param_auths=param_auths,
            today=today,
        )
    else:
        Read.create_readers(
            "",
            fixtures,
            "",
            save=False,
            use_saved=True,
            user_agent="test",
            header_auths=header_auths,
            basic_auths=basic_auths,
            param_auths=param_auths,
            today=today,
        )

    return Configuration.read()


@pytest.fixture(scope="session")
def fallbacks(fixtures):
    Fallbacks.add(join(fixtures, "fallbacks.json"), sources_key="sources")


def check_scrapers(
    names,
    runner,
    level_name,
    headers,
    values,
    sources,
    population_lookup=None,
    fallbacks_used=False,
    source_urls=None,
):
    for name in names:
        scraper = runner.get_scraper(name)
        bool_assert(scraper.has_run, True, "Scraper has not run!")
        bool_assert(
            scraper.fallbacks_used,
            fallbacks_used,
            f"Fallbacks used {scraper.fallbacks_used} is not as expected!",
        )
    results = runner.get_results(names)[level_name]
    assert results["headers"] == headers
    assert results["values"] == values
    assert results["sources"] == sources
    if population_lookup is not None:
        assert BaseScraper.population_lookup == population_lookup
    if source_urls is not None:
        assert runner.get_source_urls() == source_urls


def check_scraper(
    name,
    runner,
    level_name,
    headers,
    values,
    sources,
    population_lookup=None,
    fallbacks_used=False,
    source_urls=None,
):
    check_scrapers(
        (name,),
        runner,
        level_name,
        headers,
        values,
        sources,
        population_lookup,
        fallbacks_used,
        source_urls,
    )


def run_check_scraper(
    name,
    runner,
    level_name,
    headers,
    values,
    sources,
    population_lookup=None,
    fallbacks_used=False,
    source_urls=None,
    set_not_run=True,
):
    runner.run_one(name)
    check_scraper(
        name,
        runner,
        level_name,
        headers,
        values,
        sources,
        population_lookup,
        fallbacks_used,
        source_urls,
    )
    if set_not_run:
        runner.set_not_run(name)


def run_check_scrapers(
    names,
    runner,
    level_name,
    headers,
    values,
    sources,
    population_lookup=None,
    fallbacks_used=False,
    source_urls=None,
    set_not_run=True,
):
    runner.run(names)
    check_scrapers(
        names,
        runner,
        level_name,
        headers,
        values,
        sources,
        population_lookup,
        fallbacks_used,
        source_urls,
    )
    if set_not_run:
        runner.set_not_run_many(names)



from copy import deepcopy

from hdx.location.adminone import AdminOne
from hdx.utilities.dateparse import parse_date

from hdx.scraper.base_scraper import BaseScraper
from hdx.scraper.runner import Runner

from .conftest import check_scraper, check_scrapers
from .education_closures import EducationClosures
from .education_enrolment import EducationEnrolment


class TestScrapersCustom:
    def test_get_custom(self, configuration, fallbacks):
        BaseScraper.population_lookup = dict()
        source_date = "2022-04-30"
        today = parse_date("2020-10-01")
        adminone = AdminOne(configuration)
        level = "national"
        countries = ("AFG",)

        class Region:
            iso3_to_region_and_hrp = {"AFG": ("ROAP",)}

        region = Region()
        runner = Runner(("AFG",), adminone, today)
        datasetinfo = configuration["education_closures"]
        education_closures = EducationClosures(
            datasetinfo, today, countries, region
        )
        runner.add_custom(education_closures)
        runner.run()
        name = education_closures.name
        headers = (["School Closure"], ["#impact+type"])
        values = [{"AFG": "Closed due to COVID-19"}]
        sources = [
            (
                "#impact+type",
                source_date,
                "UNESCO",
                "https://data.humdata.org/dataset/global-school-closures-covid19",
            )
        ]
        check_scraper(name, runner, "national", headers, values, sources)
        headers = (["No. closed countries"], ["#status+country+closed"])
        values = [{"ROAP": 1}]
        sources = [
            (
                "#status+country+closed",
                source_date,
                "UNESCO",
                "https://data.humdata.org/dataset/global-school-closures-covid19",
            )
        ]
        check_scraper(name, runner, "regional", headers, values, sources)

        scraper_configuration = configuration[f"scraper_{level}"]
        runner.add_configurables(scraper_configuration, level)
        runner.run_one("population")
        check_scraper(name, runner, "regional", headers, values, sources)
        names = ("population", name)
        assert (
            runner.get_headers(
                names=names,
                levels=("regional",),
                hxltags=("#population", "#status+country+closed"),
            )["regional"]
            == headers
        )
        headers = (
            ["Population", "School Closure"],
            ["#population", "#impact+type"],
        )
        values = [{"AFG": 38041754}, {"AFG": "Closed due to COVID-19"}]
        sources = [
            (
                "#population",
                "2020-10-01",
                "World Bank",
                "https://data.humdata.org/organization/world-bank-group",
            ),
            (
                "#impact+type",
                source_date,
                "UNESCO",
                "https://data.humdata.org/dataset/global-school-closures-covid19",
            ),
        ]
        assert (
            runner.get_headers(
                names=names,
                levels=("national",),
                headers=("Population", "School Closure"),
            )["national"]
            == headers
        )
        check_scrapers(
            names,
            runner,
            "national",
            headers,
            values,
            sources,
        )
        assert runner.get_source_urls() == [
            "https://data.humdata.org/dataset/global-school-closures-covid19",
            "https://data.humdata.org/organization/world-bank-group",
        ]

        runner = Runner(("AFG",), adminone, today)
        datasetinfo = deepcopy(datasetinfo)
        datasetinfo["url"] = "NOTEXIST.csv"
        education_closures = EducationClosures(
            datasetinfo, today, countries, Region()
        )
        runner.add_custom(education_closures)
        runner.run()
        name = education_closures.name
        headers = (["School Closure"], ["#impact+type"])
        values = [{"AFG": "Closed due to COVID-19"}]
        sources = [
            (
                "#impact+type",
                "2020-09-01",
                "UNESCO",
                "tests/fixtures/fallbacks.json",
            )
        ]
        check_scraper(
            name,
            runner,
            "national",
            headers,
            values,
            sources,
            fallbacks_used=True,
        )
        headers = (["No. closed countries"], ["#status+country+closed"])
        values = [{"ROAP": 3}]
        sources = [
            (
                "#status+country+closed",
                "2020-09-01",
                "UNESCO",
                "tests/fixtures/fallbacks.json",
            )
        ]
        check_scraper(
            name,
            runner,
            "regional",
            headers,
            values,
            sources,
            fallbacks_used=True,
        )

        datasetinfo = configuration["education_enrolment"]
        education_enrolment = EducationEnrolment(
            datasetinfo, education_closures, countries, region
        )
        runner.add_custom(education_enrolment)
        runner.run()
        name = education_enrolment.name
        headers = (
            [
                "No. pre-primary to upper-secondary learners",
                "No. tertiary learners",
                "No. affected learners",
            ],
            [
                "#population+learners+pre_primary_to_secondary",
                "#population+learners+tertiary",
                "#affected+learners",
            ],
        )
        values = [{"AFG": 9865894}, {"AFG": 430980}, {"AFG": 10296874}]
        sources = [
            (
                "#population+learners+pre_primary_to_secondary",
                "2022-04-30",
                "UNESCO",
                "https://data.humdata.org/dataset/global-school-closures-covid19",
            ),
            (
                "#population+learners+tertiary",
                "2022-04-30",
                "UNESCO",
                "https://data.humdata.org/dataset/global-school-closures-covid19",
            ),
            (
                "#affected+learners",
                "2022-04-30",
                "UNESCO",
                "https://data.humdata.org/dataset/global-school-closures-covid19",
            ),
        ]
        check_scraper(name, runner, "national", headers, values, sources)
        headers = (
            ["No. affected learners", "Percentage affected learners"],
            ["#affected+learners", "#affected+learners+pct"],
        )
        values = [{"ROAP": 10296874}, {"ROAP": "1.0000"}]
        sources = [
            (
                "#affected+learners",
                "2022-04-30",
                "UNESCO",
                "https://data.humdata.org/dataset/global-school-closures-covid19",
            ),
            (
                "#affected+learners+pct",
                "2022-04-30",
                "UNESCO",
                "https://data.humdata.org/dataset/global-school-closures-covid19",
            ),
        ]
        check_scraper(name, runner, "regional", headers, values, sources)

        # closures used fallbacks so there is no no url from that scraper
        assert runner.get_source_urls() == [
            "https://data.humdata.org/dataset/global-school-closures-covid19"
        ]

        datasetinfo = configuration["education_closures"]
        education_closures = EducationClosures(
            datasetinfo, today, countries, region
        )
        runner.add_custom(education_closures)
        runner.run()
        assert runner.get_source_urls() == [
            "https://data.humdata.org/dataset/global-school-closures-covid19",
        ]

        assert runner.get_scraper_names() == [
            "education_closures",
            "education_enrolment",
        ]
        runner.prioritise_scrapers(("education_enrolment",))
        assert runner.get_scraper_names() == [
            "education_enrolment",
            "education_closures",
        ]

        education_closures.add_hxltag_source("test", "#lala")
        sources = education_closures.get_sources("test")
        assert sources == [
            (
                "#lala",
                "2022-04-30",
                "UNESCO",
                "https://data.humdata.org/dataset/global-school-closures-covid19",
            )
        ]



import copy
import logging
import sys
from typing import Dict, List, Tuple, Union

from hdx.utilities.dictandlist import dict_of_lists_add
from hdx.utilities.text import (  # noqa: F401
    get_fraction_str,
    get_numeric_if_possible,
    number_format,
)
from slugify import slugify

from hdx.scraper.base_scraper import BaseScraper

logger = logging.getLogger(__name__)


class Aggregator(BaseScraper):
    def __init__(
        self,
        name: str,
        datasetinfo: Dict,
        headers: Dict[str, Tuple],
        adm_aggregation: Union[Dict, List],
        input_values: Dict,
        aggregation_scrapers: [List[BaseScraper]],
        use_hxl: bool,
    ):
        super().__init__(name, datasetinfo, headers)
        self.adm_aggregation = adm_aggregation
        self.input_values = input_values
        self.aggregation_scrapers = aggregation_scrapers
        self.use_hxl = use_hxl

    @classmethod
    def get_scrapers(
        cls,
        configuration,
        input_level,
        output_level,
        adm_aggregation,
        runner,
        use_hxl=True,
    ):
        input_results = runner.get_results(levels=input_level, has_run=False)
        aggregation_scrapers = list()
        if not input_results:
            return aggregation_scrapers
        if use_hxl:
            main_index = 1
            header_or_hxltag_str = "hxltag"
        else:
            main_index = 0
            header_or_hxltag_str = "header"
        input_results = input_results[input_level]
        input_headers = input_results["headers"]
        input_vals = input_results["values"]
        input_values = dict()
        for index, input_header_or_hxltag in enumerate(
            input_headers[main_index]
        ):
            input_values[input_header_or_hxltag] = input_vals[index]
        if not isinstance(adm_aggregation, dict):
            adm_aggregation = {x: ("value",) for x in adm_aggregation}
        for header_or_hxltag, process_info in configuration.items():
            process_info = copy.deepcopy(process_info)
            name = f"{slugify(header_or_hxltag.lower(), separator='_')}_{output_level}"
            config_headers_or_hxltags = process_info.get("input")
            if config_headers_or_hxltags:
                exists = True
                for i, config_header_or_hxltag in enumerate(
                    config_headers_or_hxltags
                ):
                    try:
                        input_headers[main_index].index(
                            config_header_or_hxltag
                        )
                    except ValueError:
                        logger.error(
                            f"{output_level} {header_or_hxltag_str} {header_or_hxltag} not found in {input_level} input!"
                        )
                        exists = False
                        break
                if not exists:
                    continue
                output = process_info["output"]
                if use_hxl:
                    headers = ((output,), (header_or_hxltag,))
                else:
                    headers = ((header_or_hxltag,), (output,))
                scraper = cls(
                    name,
                    process_info,
                    {output_level: headers},
                    adm_aggregation,
                    input_values,
                    aggregation_scrapers,
                    use_hxl,
                )
                aggregation_scrapers.append(scraper)
            else:
                try:
                    index = input_headers[main_index].index(header_or_hxltag)
                    process_info["input"] = (header_or_hxltag,)
                    header_or_hxltag = process_info.get(
                        "rename", header_or_hxltag
                    )
                    if use_hxl:
                        output = process_info.get(
                            "output", input_headers[0][index]
                        )
                        headers = (
                            (output,),
                            (header_or_hxltag,),
                        )
                    else:
                        output = process_info.get(
                            "output", input_headers[1][index]
                        )
                        headers = (
                            (header_or_hxltag,),
                            (output,),
                        )
                    scraper = cls(
                        name,
                        process_info,
                        {output_level: headers},
                        adm_aggregation,
                        input_values,
                        aggregation_scrapers,
                        use_hxl,
                    )
                    aggregation_scrapers.append(scraper)
                except ValueError:
                    logger.error(
                        f"{output_level} {header_or_hxltag_str} {header_or_hxltag} not found in {input_level} input!"
                    )
        return aggregation_scrapers

    @staticmethod
    def get_float_or_int(valuestr):
        if not valuestr or valuestr == "N/A":
            return None
        if "." in valuestr:
            return float(valuestr)
        else:
            return int(valuestr)

    @classmethod
    def get_numeric(cls, valuestr):
        if isinstance(valuestr, str):
            total = 0
            hasvalues = False
            for value in valuestr.split("|"):
                value = cls.get_float_or_int(value)
                if value:
                    hasvalues = True
                    total += value
            if hasvalues is False:
                return ""
            return total
        return valuestr

    def process(self, output_level: str, output_hxltag, output_values):
        population_key = self.datasetinfo.get("population_key")

        def add_population(adm, value):
            if not value:
                return
            if output_hxltag == "#population":
                if population_key is None:
                    self.population_lookup[adm] = value
                else:
                    self.population_lookup[population_key] = value

        action = self.datasetinfo["action"]
        if action == "sum" or action == "mean":
            for output_adm, valuelist in output_values.items():
                total = ""
                novals = 0
                for valuestr in valuelist:
                    value = ""
                    if isinstance(valuestr, int) or isinstance(
                        valuestr, float
                    ):
                        value = valuestr
                    else:
                        if valuestr:
                            value = self.get_numeric(valuestr)
                    if value != "":
                        novals += 1
                        if total == "":
                            total = value
                        else:
                            total += value
                if action == "mean":
                    if not isinstance(total, str):
                        if isinstance(total, int):
                            quotient, remainder = divmod(total, novals)
                            if remainder == 0:
                                total = quotient
                            else:
                                total /= novals
                        else:
                            total /= novals
                if isinstance(total, float):
                    output_values[output_adm] = number_format(
                        total, trailing_zeros=False
                    )
                else:
                    output_values[output_adm] = total
                    add_population(output_adm, total)
        elif action == "range":
            for output_adm, valuelist in output_values.items():
                min = sys.maxsize
                max = -min
                for valuestr in valuelist:
                    if valuestr:
                        value = self.get_numeric(valuestr)
                        if value > max:
                            max = value
                        if value < min:
                            min = value
                if min == sys.maxsize or max == -sys.maxsize:
                    output_values[output_adm] = ""
                else:
                    if isinstance(max, float):
                        max = number_format(max, trailing_zeros=False)
                    if isinstance(min, float):
                        min = number_format(min, trailing_zeros=False)
                    output_values[output_adm] = f"{str(min)}-{str(max)}"
        elif action == "eval":
            formula = self.datasetinfo["formula"]
            if population_key is None:
                population_str = "self.population_lookup[output_adm]"
            else:
                population_str = "self.population_lookup[population_key]"
            arbitrary_string = "#pzbgvjh"
            headers_or_hxltags = list()
            if self.use_hxl:
                index = 1
            else:
                index = 0
            for aggregation_scraper in self.aggregation_scrapers:
                headers_or_hxltags.append(
                    aggregation_scraper.get_headers(output_level)[index][0]
                )
            # Indices of list sorted by length
            sorted_len_indices = sorted(
                range(len(headers_or_hxltags)),
                key=lambda k: len(headers_or_hxltags[k]),
                reverse=True,
            )
            for output_adm, valuelist in output_values.items():
                toeval = formula.replace("#population", arbitrary_string)
                for i in sorted_len_indices:
                    aggregation_scraper = self.aggregation_scrapers[i]
                    values = aggregation_scraper.get_values(output_level)[0]
                    value = values.get(output_adm, "")
                    toeval = toeval.replace(headers_or_hxltags[i], str(value))
                toeval = toeval.replace(arbitrary_string, population_str)
                total = eval(toeval)
                output_values[output_adm] = total
                add_population(output_adm, total)

    def run(self):
        output_level = next(iter(self.headers.keys()))
        output_headers = self.get_headers(output_level)
        output_valdicts = self.get_values(output_level)
        output_values = output_valdicts[0]
        input_valdicts = list()
        input_headers_or_hxltags = self.datasetinfo["input"]
        for input_header_or_hxltag in input_headers_or_hxltags:
            input_valdicts.append(self.input_values[input_header_or_hxltag])
        found_adms = set()
        for input_values in input_valdicts:
            for input_adm in input_values:
                for output_adm in self.adm_aggregation[input_adm]:
                    key = f"{output_adm}|{input_adm}"
                    if key in found_adms:
                        continue
                    value = input_values[input_adm]
                    if value is not None:
                        found_adms.add(key)
                        dict_of_lists_add(output_values, output_adm, value)
        self.process(output_level, output_headers[1][0], output_values)

    def add_sources(self):
        pass

    def add_population(self):
        pass
